class EoqErrorTypes:
    SYN = 'SYN' #Syntax error
    SEM = 'SEM' #Semantic error
    CMD = 'CMD' #Command error
    QRY = 'QRY' #Query error
    IDX = 'IDX' #Index error, e.g. index does not exist
    TYP = 'TYP' #Type error, e.g. type mismatch
    ACC = 'ACC' #Access error, access denied
    VAL = 'VAL' #Validation error
    CER = 'CER' #Call error
    CAB = 'CAL' #Aborted
    
    
# BASIC ERROR CLASS
class EoqError(Exception):
    def __init__(self,info,msg,trace=""):
        self.info = info
        self.msg = msg
        self.trace = trace
        
    def __str__(self):
        return str(self.info)+"::"+self.msg+(("::TRACE:"+self.trace) if self.trace else "")
    
    
# SPECIALIZED ERROR CLASS
class EoqSynError(EoqError):
    def __init__(self,msg,trace=""):
        super().__init__(EoqErrorTypes.SYN, msg, trace)
        
class EoqSemError(EoqError):
    def __init__(self,msg,trace=""):
        super().__init__(EoqErrorTypes.SEM, msg, trace)
        
class EoqCallError(EoqError):
    def __init__(self,msg,trace=""):
        super().__init__(EoqErrorTypes.CER, msg, trace)
        
class EoqCallAborted(EoqError):
    def __init__(self,msg,trace=""):
        super().__init__(EoqErrorTypes.CAB, msg, trace)



'''
 Bjoern Annighoefer 2019
'''

from .serializer import Serializer

from ..query.query import QrySegTypes
from ..command.command import CmdTypes
from ..util.error import EoqError



'''
Javascript Serializer   

'''

JS_PREFIX = "new eoq2."
   
class JsSerializer(Serializer):
    def __init__(self):
        self.cmdTranslators = {
            CmdTypes.GET : lambda o,p: p+"Get("+self._Ser(o.a,p)+")",
            CmdTypes.SET : lambda o,p: p+"Set("+",".join([self._Ser(a,p) for a in o.a])+")",
            CmdTypes.ADD : lambda o,p: p+"Add("+",".join([self._Ser(a,p) for a in o.a])+")",
            CmdTypes.REM : lambda o,p: p+"Rem("+",".join([self._Ser(a,p) for a in o.a])+")",
            CmdTypes.MOV : lambda o,p: p+"Mov("+",".join([self._Ser(a,p) for a in o.a])+")",
            CmdTypes.CLO : lambda o,p: p+"Clo("+",".join([self._Ser(a,p) for a in o.a])+")",
            CmdTypes.CRN : lambda o,p: p+"Crn("+",".join([self._Ser(a,p) for a in o.a])+")",
            CmdTypes.CRT : lambda o,p: p+"Crt("+",".join([self._Ser(a,p) for a in o.a])+")",
            CmdTypes.QRF : lambda o,p: p+"Qrf("+self._Ser(o.a,p)+")",
            CmdTypes.CPR : lambda o,p: p+"Cpr("+self._Ser(o.a[0],p)+","+self._Ser(o.a[1],p)+","+self._Ser(o.a[2],p)+")",
            CmdTypes.MRG : lambda o,p: p+"Mrg("+self._Ser(o.a[0],p)+","+self._Ser(o.a[1],p)+","+self._Ser(o.a[2],p)+")",
            CmdTypes.GMM : lambda o,p: p+"Gmm()",
            CmdTypes.RMM : lambda o,p: p+"Rmm("+self._Ser(o.a,p)+")",
            CmdTypes.UMM : lambda o,p: p+"Umm("+self._Ser(o.a,p)+")",
            CmdTypes.HEL : lambda o,p: p+"Hel("+self._Ser(o.a[0],p)+","+self._Ser(o.a[1],p)+")",
            CmdTypes.SES : lambda o,p: p+"Ses("+self._Ser(o.a,p)+")",
            CmdTypes.GBY : lambda o,p: p+"Gby("+self._Ser(o.a,p)+")",
            CmdTypes.STS : lambda o,p: p+"Sts()",
            CmdTypes.CHG : lambda o,p: p+"Chg("+",".join([self._Ser(a,p) for a in o.a])+")",
            CmdTypes.OBS : lambda o,p: p+"Obs("+self._Ser(o.a[0],p)+","+self._Ser(o.a[1],p)+")",
            CmdTypes.UBS : lambda o,p: p+"Ubs("+self._Ser(o.a[0],p)+","+self._Ser(o.a[1],p)+")",
            CmdTypes.GAA : lambda o,p: p+"Gaa()",
            CmdTypes.CAL : lambda o,p: p+"Cal("+self._Ser(o.a[0],p)+","+self._Ser(o.a[1:],p)+")",
            CmdTypes.ASC : lambda o,p: p+"Asc("+self._Ser(o.a[0],p)+","+self._Ser(o.a[1:],p)+")",
            CmdTypes.ABC : lambda o,p: p+"Abc("+self._Ser(o.a,p)+")",
            CmdTypes.ADC : lambda o,p: p+"Adc("+",".join([self._Ser(a,p) for a in o.a])+")",
            CmdTypes.GAC : lambda o,p: p+"Gac()",
            CmdTypes.VAL : lambda o,p: p+"Val()",
            CmdTypes.RMC : lambda o,p: p+"Rmc("+self._Ser(o.a,p)+")",
            CmdTypes.CST : lambda o,p: p+"Cst("+self._Ser(o.a,p)+")",
            CmdTypes.CMP : lambda o,p: p+"Cmp()."+".".join([self._Ser(a,"") for a in o.a]),
            CmdTypes.MUT : lambda o,p: p+"Mut()",
            CmdTypes.UMT : lambda o,p: p+"Umt()"
            }
        self.qryTranslators = {     
            QrySegTypes.OBJ : lambda o,p: "Obj("+self._Ser(o.v,p)+")",
            QrySegTypes.HIS : lambda o,p: "His("+self._Ser(o.v,p)+")",
            QrySegTypes.PTH : lambda o,p: "Pth("+self._Ser(o.v,p)+")",
            QrySegTypes.CLS : lambda o,p: "Cls("+self._Ser(o.v,p)+")",
            QrySegTypes.INO : lambda o,p: "Ino("+self._Ser(o.v,p)+")",
            QrySegTypes.MET : lambda o,p: "Met("+self._Ser(o.v[0],p)+(","+self._Ser(o.v[1:],p) if len(o.v)>1 else "")+")",
            QrySegTypes.NOT : lambda o,p: "Not()",
            QrySegTypes.TRM : lambda o,p: "Trm("+self._Ser(o.v[0],p)+","+self._Ser(o.v[1],p)+")",
            QrySegTypes.CNS : lambda o,p: "Cns()",
            QrySegTypes.TRY : lambda o,p: "Try("+self._Ser(o.v[0],p)+","+self._Ser(o.v[1],p)+")",
            QrySegTypes.IDX : lambda o,p: "Idx("+self._Ser(o.v,p)+")",
            QrySegTypes.SEL : lambda o,p: "Sel("+self._Ser(o.v,p)+")",
            QrySegTypes.ARR : lambda o,p: "Arr(["+",".join([self._Ser(a,p) for a in o.v])+"])",
            QrySegTypes.ZIP : lambda o,p: "Zip("+self._Ser(o.v,p)+")",
            QrySegTypes.QRY : lambda o,p: p+"Qry()."+".".join([self._Ser(a,p) for a in o.v]),
            QrySegTypes.ANY : lambda o,p: "Any("+self._Ser(o.v,p)+")",
            QrySegTypes.ALL : lambda o,p: "All("+self._Ser(o.v,p)+")",
            QrySegTypes.EQU : lambda o,p: "Equ("+self._Ser(o.v,p)+")",
            QrySegTypes.EQA : lambda o,p: "Eqa("+self._Ser(o.v,p)+")",
            QrySegTypes.NEQ : lambda o,p: "Neq("+self._Ser(o.v,p)+")",
            QrySegTypes.LES : lambda o,p: "Les("+self._Ser(o.v,p)+")",
            QrySegTypes.GRE : lambda o,p: "Gre("+self._Ser(o.v,p)+")",
            QrySegTypes.RGX : lambda o,p: "Rgx("+self._Ser(o.v,p)+")",
            QrySegTypes.ADD : lambda o,p: "Add("+self._Ser(o.v,p)+")",
            QrySegTypes.SUB : lambda o,p: "Sub("+self._Ser(o.v,p)+")",
            QrySegTypes.MUL : lambda o,p: "Mul("+self._Ser(o.v,p)+")",
            QrySegTypes.ORR : lambda o,p: "Orr("+self._Ser(o.v,p)+")",
            QrySegTypes.XOR : lambda o,p: "Xor("+self._Ser(o.v,p)+")",
            QrySegTypes.AND : lambda o,p: "And("+self._Ser(o.v,p)+")",
            QrySegTypes.NAD : lambda o,p: "Nad("+self._Ser(o.v,p)+")",
            QrySegTypes.DIV : lambda o,p: "Div("+self._Ser(o.v,p)+")",
            QrySegTypes.CSP : lambda o,p: "Csp("+self._Ser(o.v,p)+")",
            QrySegTypes.ITS : lambda o,p: "Its("+self._Ser(o.v,p)+")",
            QrySegTypes.DIF : lambda o,p: "Dif("+self._Ser(o.v,p)+")",
            QrySegTypes.UNI : lambda o,p: "Uni("+self._Ser(o.v,p)+")",
            QrySegTypes.CON : lambda o,p: "Con("+self._Ser(o.v,p)+")"
            }
        self.priTranslators = {
            #primitive types
            bool : lambda o,p: str(o),
            int : lambda o,p: str(o),
            float : lambda o,p: str(o),
            bool : lambda o,p: str(o),
            str : lambda o,p: "'"+o+"'",
            list : lambda o,p: "[%s]"%(",".join([self._Ser(x,p) for x in o])),
            type(None) : lambda o,p: str(o)
            }
        
    def Ser(self,val):
        return self._Ser(val,JS_PREFIX)
    
    def _Ser(self,val,p):
        try: 
            return self.cmdTranslators[val.cmd](val,p)
        except:
            try: 
                return self.qryTranslators[val.qry](val,JS_PREFIX)
            except:
                try:
                    return self.priTranslators[type(val)](val,JS_PREFIX)
                except:
                    raise EoqError(0,"Text serializer failed for %s"%(str(val)))
    
    def Des(self,code):
        pass


'''
 2019 Bjoern Annighoefer
'''

'''
 CALL HANDLER
'''
    
class CallHandler:
    def __init__(self):
        pass
    
    def CanHandle(self,name,args,ctype):
        pass
        
    def HandleSync(self,name,callId,domain,args,opts,callManager):
        pass
    
    def HandleAsync(self,name,callId,domain,args,opts,callManager,sessionId=None):
        pass
    
    def InputToCall(self,callId,channelName,data):
        pass
    
    def AbortCall(self,callId):
        pass
    
    
class FunctionWrapperCallHandler(CallHandler):
    def __init__(self,fct):
        self.fct = fct
        
    def HandleSync(self,name,callId,domain,args,opts,callManager,sessionId=None):
        return self.fct(domain,*args)


from pyecore.resources import global_registry
from .xmlresourcemodel import getEClassifier, eClassifiers
from .xmlresourcemodel import name, nsURI, nsPrefix, eClass
from .xmlresourcemodel import Document, Element, Attribute, CommentA, Comment, CData


from . import xmlresourcemodel

__all__ = ['Document', 'Element', 'Attribute', 'CommentA', 'Comment', 'CData']

eSubpackages = []
eSuperPackage = None
xmlresourcemodel.eSubpackages = eSubpackages
xmlresourcemodel.eSuperPackage = eSuperPackage

Document.rootelement.eType = Element
Document.comments.eType = CommentA
Element.attributes.eType = Attribute
Element.subelements.eType = Element

otherClassifiers = []

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [xmlresourcemodel] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack



"""Definition of meta model 'model'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *


name = 'model'
nsURI = 'http://www.eoq.de/model/v1.0'
nsPrefix = 'de.eoq'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)
MetaModelSourcesE = EEnum('MetaModelSourcesE', literals=['BUILDIN', 'DYNAMIC'])

FileResourceTypesE = EEnum('FileResourceTypesE', literals=['MODEL', 'TEXT', 'RAW'])

SegmentTypesE = EEnum('SegmentTypesE', literals=[
                      'RESOURCE', 'PATH', 'ID', 'CLAZZ', 'INSTANCE', 'META', 'HISTORY', 'LISTOP', 'SELECTOR', 'INDEX'])

MetaSegmentIdentifiersE = EEnum('MetaSegmentIdentifiersE', literals=['CONTAINER', 'INDEX', 'TYPE', 'CLASS', 'CONTENTS', 'CLASSNAME', 'PACKAGE', 'PACKAGENAME', 'RESOURCE', 'RESOURCENAME', 'CONTAININGFEATURE', 'CONTAININGFEATURENAME',
                                                                     'FEATURES', 'FEATURENAMES', 'ATTRIBUTES', 'ATTRIBUTENAMES', 'REFERENCES', 'REFERENCENAMES', 'CONTAINMENTS', 'CONTAINMENTNAMES', 'ATTRIBUTEVALUES', 'FEATUREVALUES', 'REFERENCEVALUES', 'CONTAINMENTVALUES'])

IndexTypesE = EEnum('IndexTypesE', literals=['NUMBER', 'RANGE', 'ADD', 'REMOVE', 'FLATTEN'])

OperatorTypesE = EEnum('OperatorTypesE', literals=['EQUAL', 'NOTEQUAL', 'GREATER', 'LESS'])

ValueTypesE = EEnum('ValueTypesE', literals=[
                    'INT', 'FLOAT', 'BOOL', 'STRING', 'OBJECTREF', 'EMPTY', 'LIST', 'HISTORYREF', 'OPERATION', 'OBJECT', 'QUERY'])

ReturnMultiplicityTypeE = EEnum('ReturnMultiplicityTypeE', literals=[
                                'UNCHANGED', 'FLATTENED', 'FORCESINGLE'])

CommandTypesE = EEnum('CommandTypesE', literals=['HELLO', 'GOODBYE', 'SESSION', 'STATUS', 'CHANGES', 'RETRIEVE',
                                                 'CREATE', 'UPDATE', 'CLONE', 'UNDO', 'CALL', 'ASYNCCALL', 'CALLSTATUS', 'ABORTCALL', 'COMPOUND'])

CloneModesE = EEnum('CloneModesE', literals=['CLASS', 'ATTRIBUTES', 'FULL', 'DEEP'])

ResultTypesE = EEnum('ResultTypesE', literals=['OK', 'ERROR', 'COMPOUND_OK', 'COMPOUND_ERROR'])

CallStatusE = EEnum('CallStatusE', literals=[
                    'INITIALIZING', 'RUNNING', 'WAITING', 'FINISHED', 'ABORTED', 'ERROR'])

CallChannelTypesE = EEnum('CallChannelTypesE', literals=['OUT', 'IN', 'INTERACTIVE'])

ListOpSegmentTypesE = EEnum('ListOpSegmentTypesE', literals=['SIZE', 'FLATTEN'])

SegmentProcessingTypesE = EEnum('SegmentProcessingTypesE', literals=['RECURSIVE', 'LINEAR'])


Object = EDataType('Object', instanceClassName='None')


@abstract
class NamedElementA(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, derived=False, changeable=True, default_value='UNNAMED')

    def __init__(self, *, name=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name


class Metamodel(EObject, metaclass=MetaEClass):

    source = EAttribute(eType=MetaModelSourcesE, derived=False, changeable=True)
    name = EAttribute(eType=EString, derived=False, changeable=True)
    package = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, source=None, package=None, name=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if source is not None:
            self.source = source

        if name is not None:
            self.name = name

        if package is not None:
            self.package = package


class CommandInfo(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, derived=False, changeable=True, iD=True)
    description = EAttribute(eType=EString, derived=False, changeable=True)
    parameters = EReference(ordered=True, unique=True, containment=True, upper=-1)
    results = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, name=None, parameters=None, description=None, results=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if description is not None:
            self.description = description

        if parameters:
            self.parameters.extend(parameters)

        if results:
            self.results.extend(results)


class Transaction(EObject, metaclass=MetaEClass):

    id = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    starttime = EAttribute(eType=EFloat, derived=False, changeable=True, default_value=0.0)
    endtime = EAttribute(eType=EFloat, derived=False, changeable=True, default_value=0.0)
    deadline = EAttribute(eType=EFloat, derived=False, changeable=True, default_value=0.0)
    maxDuration = EAttribute(eType=EFloat, derived=False, changeable=True, default_value=0.0)
    wasTimedOut = EAttribute(eType=EBoolean, derived=False, changeable=True, default_value=False)
    wasEnded = EAttribute(eType=EBoolean, derived=False, changeable=True)
    history = EReference(ordered=True, unique=True, containment=False, upper=-1)
    changes = EReference(ordered=True, unique=True, containment=False, upper=-1)

    def __init__(self, *, id=None, starttime=None, endtime=None, deadline=None, maxDuration=None, wasTimedOut=None, wasEnded=None, history=None, changes=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if id is not None:
            self.id = id

        if starttime is not None:
            self.starttime = starttime

        if endtime is not None:
            self.endtime = endtime

        if deadline is not None:
            self.deadline = deadline

        if maxDuration is not None:
            self.maxDuration = maxDuration

        if wasTimedOut is not None:
            self.wasTimedOut = wasTimedOut

        if wasEnded is not None:
            self.wasEnded = wasEnded

        if history:
            self.history.extend(history)

        if changes:
            self.changes.extend(changes)


class Change(EObject, metaclass=MetaEClass):

    changeId = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    sourceTranscallId = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    target = EReference(ordered=True, unique=True, containment=True)
    query = EReference(ordered=True, unique=True, containment=True)
    newValue = EReference(ordered=True, unique=True, containment=True)
    oldValue = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, target=None, query=None, newValue=None, changeId=None, sourceTranscallId=None, oldValue=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if changeId is not None:
            self.changeId = changeId

        if sourceTranscallId is not None:
            self.sourceTranscallId = sourceTranscallId

        if target is not None:
            self.target = target

        if query is not None:
            self.query = query

        if newValue is not None:
            self.newValue = newValue

        if oldValue is not None:
            self.oldValue = oldValue


class ActionInfo(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, derived=False, changeable=True, iD=True)
    description = EAttribute(eType=EString, derived=False, changeable=True)
    details = EAttribute(eType=EString, derived=False, changeable=True)
    parameters = EReference(ordered=True, unique=True, containment=True, upper=-1)
    results = EReference(ordered=True, unique=True, containment=True, upper=-1)
    handler = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, name=None, description=None, parameters=None, results=None, handler=None, details=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if description is not None:
            self.description = description

        if details is not None:
            self.details = details

        if parameters:
            self.parameters.extend(parameters)

        if results:
            self.results.extend(results)

        if handler is not None:
            self.handler = handler


class ActionParameter(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, derived=False, changeable=True, iD=True)
    type = EAttribute(eType=EString, derived=False, changeable=True, default_value='String')
    lowerBound = EAttribute(eType=EInt, derived=False, changeable=True, default_value=1)
    upperBound = EAttribute(eType=EInt, derived=False, changeable=True, default_value=1)
    description = EAttribute(eType=EString, derived=False, changeable=True)
    default = EAttribute(eType=EString, derived=False, changeable=True)
    choices = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, name=None, type=None, lowerBound=None, upperBound=None, description=None, default=None, choices=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if type is not None:
            self.type = type

        if lowerBound is not None:
            self.lowerBound = lowerBound

        if upperBound is not None:
            self.upperBound = upperBound

        if description is not None:
            self.description = description

        if default is not None:
            self.default = default

        if choices:
            self.choices.extend(choices)


class Choice(EObject, metaclass=MetaEClass):

    value = EAttribute(eType=EString, derived=False, changeable=True, iD=True)

    def __init__(self, *, value=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if value is not None:
            self.value = value


@abstract
class ActionHandlerA(EObject, metaclass=MetaEClass):

    def __init__(self, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

    def Call(self, actionCall=None, transaction=None):

        raise NotImplementedError('operation Call(...) not yet implemented')

    def AsyncCall(self, actionCall=None, transaction=None):

        raise NotImplementedError('operation AsyncCall(...) not yet implemented')

    def CallStatus(self, actionCall=None, transaction=None):

        raise NotImplementedError('operation CallStatus(...) not yet implemented')

    def AbortCall(self, actionCall=None, transaction=None):

        raise NotImplementedError('operation AbortCall(...) not yet implemented')


class CommandParameter(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, derived=False, changeable=True, iD=True)
    type = EAttribute(eType=EString, derived=False, changeable=True, default_value='String')
    description = EAttribute(eType=EString, derived=False, changeable=True)
    example = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, name=None, type=None, description=None, example=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if type is not None:
            self.type = type

        if description is not None:
            self.description = description

        if example is not None:
            self.example = example


class Query(EObject, metaclass=MetaEClass):

    segments = EReference(ordered=True, unique=True, containment=True, upper=-1)
    sourceClass = EReference(ordered=True, unique=True, containment=True)
    returnMultiplicity = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, segments=None, sourceClass=None, returnMultiplicity=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if segments:
            self.segments.extend(segments)

        if sourceClass is not None:
            self.sourceClass = sourceClass

        if returnMultiplicity is not None:
            self.returnMultiplicity = returnMultiplicity


@abstract
class Segment(EObject, metaclass=MetaEClass):

    identifier = EAttribute(eType=EString, derived=False, changeable=True)
    selector = EReference(ordered=True, unique=True, containment=True)
    index = EReference(ordered=True, unique=True, containment=True)
    depth = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, identifier=None, selector=None, index=None, depth=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if identifier is not None:
            self.identifier = identifier

        if selector is not None:
            self.selector = selector

        if index is not None:
            self.index = index

        if depth is not None:
            self.depth = depth


class Depth(EObject, metaclass=MetaEClass):

    value = EAttribute(eType=EInt, derived=False, changeable=True, default_value=1)

    def __init__(self, *, value=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if value is not None:
            self.value = value


@abstract
class Index(EObject, metaclass=MetaEClass):

    def __init__(self, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


class Selector(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, derived=False, changeable=True)
    value = EReference(ordered=True, unique=True, containment=False)
    operator = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, name=None, value=None, operator=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if value is not None:
            self.value = value

        if operator is not None:
            self.operator = operator


@abstract
class Operator(EObject, metaclass=MetaEClass):

    def __init__(self, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


@abstract
class Value(EObject, metaclass=MetaEClass):

    def __init__(self, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


class SourceClass(EObject, metaclass=MetaEClass):

    dontCare = EAttribute(eType=EBoolean, derived=False, changeable=True, default_value=True)
    name = EAttribute(eType=EString, derived=False, changeable=True, default_value='*')
    type = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, dontCare=None, name=None, type=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if dontCare is not None:
            self.dontCare = dontCare

        if name is not None:
            self.name = name

        if type is not None:
            self.type = type


@abstract
class ReturnMultiplicity(EObject, metaclass=MetaEClass):

    def __init__(self, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


@abstract
class CommandA(EObject, metaclass=MetaEClass):

    def __init__(self, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()


@abstract
class ResultA(EObject, metaclass=MetaEClass):

    transactionId = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)

    def __init__(self, *, transactionId=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if transactionId is not None:
            self.transactionId = transactionId


class ActionCall(EObject, metaclass=MetaEClass):

    callId = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    callStatus = EAttribute(eType=CallStatusE, derived=False, changeable=True,
                            default_value=CallStatusE.INITIALIZING)
    action = EAttribute(eType=EString, derived=False, changeable=True, iD=True)
    channels = EReference(ordered=True, unique=True, containment=True, upper=-1)
    args = EReference(ordered=True, unique=True, containment=True)
    handler = EReference(ordered=True, unique=True, containment=False)
    returnValues = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, callId=None, callStatus=None, action=None, channels=None, args=None, handler=None, returnValues=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if callId is not None:
            self.callId = callId

        if callStatus is not None:
            self.callStatus = callStatus

        if action is not None:
            self.action = action

        if channels:
            self.channels.extend(channels)

        if args is not None:
            self.args = args

        if handler is not None:
            self.handler = handler

        if returnValues is not None:
            self.returnValues = returnValues


class CallChannel(EObject, metaclass=MetaEClass):

    type = EAttribute(eType=CallChannelTypesE, derived=False,
                      changeable=True, default_value=CallChannelTypesE.OUT)
    name = EAttribute(eType=EString, derived=False, changeable=True, iD=True)
    data = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, type=None, name=None, data=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if type is not None:
            self.type = type

        if name is not None:
            self.name = name

        if data:
            self.data.extend(data)


class CallChannelData(EObject, metaclass=MetaEClass):

    date = EAttribute(eType=EString, derived=False, changeable=True)
    data = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, date=None, data=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if date is not None:
            self.date = date

        if data is not None:
            self.data = data


@abstract
class PathElementA(EObject, metaclass=MetaEClass):

    def __init__(self, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

    def actualPathAbs(self):

        raise NotImplementedError('operation actualPathAbs(...) not yet implemented')

    def actualPath(self):

        raise NotImplementedError('operation actualPath(...) not yet implemented')

    def actualPathCwd(self):

        raise NotImplementedError('operation actualPathCwd(...) not yet implemented')


@abstract
class DomainA(NamedElementA):

    version = EAttribute(eType=EString, derived=False, changeable=True, default_value='UNKNOWN')
    transactionCount = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    changeCount = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    callCount = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    metamodels = EReference(ordered=True, unique=True, containment=True, upper=-1)
    transactions = EReference(ordered=True, unique=True, containment=True, upper=-1)
    commands = EReference(ordered=True, unique=True, containment=True, upper=-1)
    models = EReference(ordered=True, unique=True, containment=True, upper=-1)
    actions = EReference(ordered=True, unique=True, containment=False, upper=-1)
    currentTransaction = EReference(ordered=True, unique=True, containment=False)
    changes = EReference(ordered=True, unique=True, containment=True, upper=-1)
    actionCalls = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, version=None, metamodels=None, transactions=None, commands=None, models=None, actions=None, currentTransaction=None, transactionCount=None, changeCount=None, changes=None, actionCalls=None, callCount=None, **kwargs):

        super().__init__(**kwargs)

        if version is not None:
            self.version = version

        if transactionCount is not None:
            self.transactionCount = transactionCount

        if changeCount is not None:
            self.changeCount = changeCount

        if callCount is not None:
            self.callCount = callCount

        if metamodels:
            self.metamodels.extend(metamodels)

        if transactions:
            self.transactions.extend(transactions)

        if commands:
            self.commands.extend(commands)

        if models:
            self.models.extend(models)

        if actions:
            self.actions.extend(actions)

        if currentTransaction is not None:
            self.currentTransaction = currentTransaction

        if changes:
            self.changes.extend(changes)

        if actionCalls:
            self.actionCalls.extend(actionCalls)

    def Do(self, command=None):

        raise NotImplementedError('operation Do(...) not yet implemented')


@abstract
class FileResourceA(PathElementA):

    name = EAttribute(eType=EString, derived=False, changeable=True, default_value='newResource')
    isDirty = EAttribute(eType=EBoolean, derived=False, changeable=True)
    isLoaded = EAttribute(eType=EBoolean, derived=False, changeable=True, default_value=False)
    lastPersistentPath = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, name=None, isDirty=None, isLoaded=None, lastPersistentPath=None, **kwargs):

        super().__init__(**kwargs)

        if name is not None:
            self.name = name

        if isDirty is not None:
            self.isDirty = isDirty

        if isLoaded is not None:
            self.isLoaded = isLoaded

        if lastPersistentPath is not None:
            self.lastPersistentPath = lastPersistentPath


class NumberIndex(Index):

    type = EAttribute(eType=IndexTypesE, derived=False,
                      changeable=True, default_value=IndexTypesE.NUMBER)
    value = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)

    def __init__(self, *, type=None, value=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if value is not None:
            self.value = value


class RangeIndex(Index):

    type = EAttribute(eType=IndexTypesE, derived=False,
                      changeable=True, default_value=IndexTypesE.RANGE)
    lower = EAttribute(eType=EInt, derived=False, changeable=True, default_value=1)
    upper = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)

    def __init__(self, *, type=None, lower=None, upper=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if lower is not None:
            self.lower = lower

        if upper is not None:
            self.upper = upper


class AddIndex(Index):

    type = EAttribute(eType=IndexTypesE, derived=False,
                      changeable=True, default_value=IndexTypesE.ADD)

    def __init__(self, *, type=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type


class RemoveIndex(Index):

    type = EAttribute(eType=IndexTypesE, derived=False,
                      changeable=True, default_value=IndexTypesE.REMOVE)

    def __init__(self, *, type=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type


class FlattenIndex(Index):

    type = EAttribute(eType=IndexTypesE, derived=False, changeable=True,
                      default_value=IndexTypesE.FLATTEN)

    def __init__(self, *, type=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type


class EqualOperator(Operator):

    type = EAttribute(eType=OperatorTypesE, derived=False,
                      changeable=True, default_value=OperatorTypesE.EQUAL)
    symbol = EAttribute(eType=EChar, derived=False, changeable=True, default_value='=')

    def __init__(self, *, type=None, symbol=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if symbol is not None:
            self.symbol = symbol


class NotEqualOperator(Operator):

    type = EAttribute(eType=OperatorTypesE, derived=False, changeable=True,
                      default_value=OperatorTypesE.NOTEQUAL)
    symbol = EAttribute(eType=EChar, derived=False, changeable=True, default_value='~')

    def __init__(self, *, type=None, symbol=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if symbol is not None:
            self.symbol = symbol


class GreaterOperator(Operator):

    type = EAttribute(eType=OperatorTypesE, derived=False, changeable=True,
                      default_value=OperatorTypesE.GREATER)
    symbol = EAttribute(eType=EChar, derived=False, changeable=True, default_value='>')

    def __init__(self, *, type=None, symbol=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if symbol is not None:
            self.symbol = symbol


class LessOperator(Operator):

    type = EAttribute(eType=OperatorTypesE, derived=False,
                      changeable=True, default_value=OperatorTypesE.LESS)
    symbol = EAttribute(eType=EChar, derived=False, changeable=True, default_value='<')

    def __init__(self, *, type=None, symbol=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if symbol is not None:
            self.symbol = symbol


class IntValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False,
                      changeable=True, default_value=ValueTypesE.INT)
    v = EAttribute(eType=EInt, derived=False, changeable=True)

    def __init__(self, *, type=None, v=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v is not None:
            self.v = v


class FloatValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False,
                      changeable=True, default_value=ValueTypesE.FLOAT)
    v = EAttribute(eType=EFloat, derived=False, changeable=True)

    def __init__(self, *, type=None, v=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v is not None:
            self.v = v


class BoolValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False,
                      changeable=True, default_value=ValueTypesE.BOOL)
    v = EAttribute(eType=EBoolean, derived=False, changeable=True)

    def __init__(self, *, type=None, v=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v is not None:
            self.v = v


class StringValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False,
                      changeable=True, default_value=ValueTypesE.STRING)
    v = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, type=None, v=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v is not None:
            self.v = v


class ObjectRefValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False, changeable=True,
                      default_value=ValueTypesE.OBJECTREF)
    v = EAttribute(eType=EInt, derived=False, changeable=True)

    def __init__(self, *, type=None, v=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v is not None:
            self.v = v


class EmptyValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False,
                      changeable=True, default_value=ValueTypesE.EMPTY)
    v = EAttribute(eType=Object, derived=False, changeable=True)

    def __init__(self, *, type=None, v=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v is not None:
            self.v = v


class ListValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False,
                      changeable=True, default_value=ValueTypesE.LIST)
    v = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, v=None, type=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v:
            self.v.extend(v)


class HistoryRefValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False, changeable=True,
                      default_value=ValueTypesE.HISTORYREF)
    v = EAttribute(eType=EInt, derived=False, changeable=True)

    def __init__(self, *, type=None, v=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v is not None:
            self.v = v


class OperationValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False, changeable=True,
                      default_value=ValueTypesE.OPERATION)

    def __init__(self, *, type=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

    def v(self):

        raise NotImplementedError('operation v(...) not yet implemented')


class ObjectValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False,
                      changeable=True, default_value=ValueTypesE.OBJECT)
    v = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, type=None, v=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v is not None:
            self.v = v


class UnchangedReturnMuliplicity(ReturnMultiplicity):

    type = EAttribute(eType=ReturnMultiplicityTypeE, derived=False,
                      changeable=True, default_value=ReturnMultiplicityTypeE.UNCHANGED)
    symbol = EAttribute(eType=EChar, derived=False, changeable=True, default_value='*')

    def __init__(self, *, type=None, symbol=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if symbol is not None:
            self.symbol = symbol


class FlattenedReturnMultiplicity(ReturnMultiplicity):

    type = EAttribute(eType=ReturnMultiplicityTypeE, derived=False,
                      changeable=True, default_value=ReturnMultiplicityTypeE.FLATTENED)
    symbol = EAttribute(eType=EChar, derived=False, changeable=True, default_value='_')

    def __init__(self, *, type=None, symbol=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if symbol is not None:
            self.symbol = symbol


class ForceSingleReturnMultiplicity(ReturnMultiplicity):

    type = EAttribute(eType=ReturnMultiplicityTypeE, derived=False, changeable=True,
                      default_value=ReturnMultiplicityTypeE.FORCESINGLE)
    symbol = EAttribute(eType=EChar, derived=False, changeable=True, default_value='1')

    def __init__(self, *, type=None, symbol=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if symbol is not None:
            self.symbol = symbol


class HelloCommand(CommandA):
    """Grants access as a certain user to the domain. Returned is a session id."""
    type = EAttribute(eType=CommandTypesE, derived=False,
                      changeable=True, default_value=CommandTypesE.HELLO)
    identification = EReference(ordered=True, unique=True, containment=True)
    sessionId = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, identification=None, sessionId=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if identification is not None:
            self.identification = identification

        if sessionId is not None:
            self.sessionId = sessionId


class GoodbyeCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False, changeable=True,
                      default_value=CommandTypesE.GOODBYE)
    sessionId = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, sessionId=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if sessionId is not None:
            self.sessionId = sessionId


class SessionCommand(CommandA):
    """When working as a user this should be the first command in a compound command to notify the domain about the session to be used. """
    type = EAttribute(eType=CommandTypesE, derived=False, changeable=True,
                      default_value=CommandTypesE.SESSION)
    sessionId = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, sessionId=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if sessionId is not None:
            self.sessionId = sessionId


class StatusCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False,
                      changeable=True, default_value=CommandTypesE.STATUS)

    def __init__(self, *, type=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type


class ChangesCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False, changeable=True,
                      default_value=CommandTypesE.CHANGES)
    earliestChangeId = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, earliestChangeId=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if earliestChangeId is not None:
            self.earliestChangeId = earliestChangeId


class RetrieveCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False, changeable=True,
                      default_value=CommandTypesE.RETRIEVE)
    target = EReference(ordered=True, unique=True, containment=True)
    query = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, type=None, target=None, query=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if target is not None:
            self.target = target

        if query is not None:
            self.query = query


class CreateCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False,
                      changeable=True, default_value=CommandTypesE.CREATE)
    packageNsUri = EReference(ordered=True, unique=True, containment=True)
    className = EReference(ordered=True, unique=True, containment=True)
    n = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, packageNsUri=None, className=None, n=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if packageNsUri is not None:
            self.packageNsUri = packageNsUri

        if className is not None:
            self.className = className

        if n is not None:
            self.n = n


class UpdateCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False,
                      changeable=True, default_value=CommandTypesE.UPDATE)
    target = EReference(ordered=True, unique=True, containment=True)
    query = EReference(ordered=True, unique=True, containment=True)
    value = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, target=None, query=None, value=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if target is not None:
            self.target = target

        if query is not None:
            self.query = query

        if value is not None:
            self.value = value


class CloneCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False,
                      changeable=True, default_value=CommandTypesE.CLONE)
    mode = EAttribute(eType=CloneModesE, derived=False, changeable=True)
    target = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, target=None, mode=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if mode is not None:
            self.mode = mode

        if target is not None:
            self.target = target


class CallCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False,
                      changeable=True, default_value=CommandTypesE.CALL)
    action = EReference(ordered=True, unique=True, containment=True)
    args = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, action=None, args=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if action is not None:
            self.action = action

        if args is not None:
            self.args = args


class AsyncCallCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False, changeable=True,
                      default_value=CommandTypesE.ASYNCCALL)
    action = EReference(ordered=True, unique=True, containment=True)
    args = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, action=None, args=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if action is not None:
            self.action = action

        if args is not None:
            self.args = args


class CallStatusCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False, changeable=True,
                      default_value=CommandTypesE.CALLSTATUS)
    callId = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, callId=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if callId is not None:
            self.callId = callId


class AbortCallCommand(CommandA):

    type = EAttribute(eType=CommandTypesE, derived=False, changeable=True,
                      default_value=CommandTypesE.ABORTCALL)
    callId = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, callId=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if callId is not None:
            self.callId = callId


class CompoundCommand(CommandA):
    """A sequence of commands executed within the same transaction and therefore being atomar."""
    type = EAttribute(eType=CommandTypesE, derived=False, changeable=True,
                      default_value=CommandTypesE.COMPOUND)
    commands = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, commands=None, type=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commands:
            self.commands.extend(commands)


class HelloResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.HELLO)
    sessionId = EAttribute(eType=EString, derived=False, changeable=True,
                           default_value='123456789ABCDEF')

    def __init__(self, *, type=None, commandType=None, sessionId=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if sessionId is not None:
            self.sessionId = sessionId


class GoodbyeResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.GOODBYE)

    def __init__(self, *, type=None, commandType=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType


class SessionResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.SESSION)

    def __init__(self, *, type=None, commandType=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType


class StatusResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.STATUS)
    changeId = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)

    def __init__(self, *, type=None, commandType=None, changeId=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if changeId is not None:
            self.changeId = changeId


class ChangesResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.CHANGES)
    changes = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, commandType=None, changes=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if changes is not None:
            self.changes = changes


class RetrieveResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.RETRIEVE)
    value = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, commandType=None, value=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if value is not None:
            self.value = value


class CreateResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.CREATE)
    value = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, commandType=None, value=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if value is not None:
            self.value = value


class UpdateResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.UPDATE)
    target = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, commandType=None, target=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if target is not None:
            self.target = target


class CloneResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.CLONE)
    value = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, commandType=None, value=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if value is not None:
            self.value = value


class CallResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.CALL)
    callId = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    returnValues = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, type=None, commandType=None, callId=None, returnValues=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if callId is not None:
            self.callId = callId

        if returnValues is not None:
            self.returnValues = returnValues


class AsyncCallResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.ASYNCCALL)
    callId = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)

    def __init__(self, *, type=None, commandType=None, callId=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if callId is not None:
            self.callId = callId


class CallStatusResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.CALLSTATUS)
    callStatus = EAttribute(eType=CallStatusE, derived=False, changeable=True)
    callId = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    result = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, commandType=None, callStatus=None, callId=None, result=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if callStatus is not None:
            self.callStatus = callStatus

        if callId is not None:
            self.callId = callId

        if result is not None:
            self.result = result


class AbortCallResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.ABORTCALL)

    def __init__(self, *, type=None, commandType=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType


class ErrorResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False,
                      changeable=True, default_value=ResultTypesE.ERROR)
    commandType = EAttribute(eType=CommandTypesE, derived=False, changeable=True)
    code = EAttribute(eType=EInt, derived=False, changeable=True, default_value=0)
    message = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, type=None, commandType=None, code=None, message=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if code is not None:
            self.code = code

        if message is not None:
            self.message = message


class CompoundResult(ResultA):

    type = EAttribute(eType=ResultTypesE, derived=False, changeable=True,
                      default_value=ResultTypesE.COMPOUND_OK)
    commandType = EAttribute(eType=CommandTypesE, derived=False,
                             changeable=True, default_value=CommandTypesE.COMPOUND)
    results = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, type=None, commandType=None, results=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if commandType is not None:
            self.commandType = commandType

        if results:
            self.results.extend(results)


class QueryValue(Value):

    type = EAttribute(eType=ValueTypesE, derived=False,
                      changeable=True, default_value=ValueTypesE.QUERY)
    v = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, type=None, v=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if v is not None:
            self.v = v


@abstract
class RecursiveSegmetA(Segment):

    processingType = EAttribute(eType=SegmentProcessingTypesE, derived=False,
                                changeable=True, default_value=SegmentProcessingTypesE.RECURSIVE)

    def __init__(self, *, processingType=None, **kwargs):

        super().__init__(**kwargs)

        if processingType is not None:
            self.processingType = processingType


@abstract
class LinearSegmetA(Segment):

    processingType = EAttribute(eType=SegmentProcessingTypesE, derived=False,
                                changeable=True, default_value=SegmentProcessingTypesE.LINEAR)

    def __init__(self, *, processingType=None, **kwargs):

        super().__init__(**kwargs)

        if processingType is not None:
            self.processingType = processingType


class Directory(NamedElementA, PathElementA):

    subdirectories = EReference(ordered=True, unique=True, containment=True, upper=-1)
    resources = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, subdirectories=None, resources=None, **kwargs):

        super().__init__(**kwargs)

        if subdirectories:
            self.subdirectories.extend(subdirectories)

        if resources:
            self.resources.extend(resources)


class ModelResource(FileResourceA):

    type = EAttribute(eType=FileResourceTypesE, derived=False,
                      changeable=True, default_value=FileResourceTypesE.MODEL)
    isMetaModel = EAttribute(eType=EBoolean, derived=False,
                             changeable=True, upper=-1, default_value=False)
    isWritable = EAttribute(eType=EBoolean, derived=False,
                            changeable=True, upper=-1, default_value=True)
    contents = EReference(ordered=True, unique=True, containment=True, upper=-1)
    domain = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, type=None, isMetaModel=None, isWritable=None, contents=None, domain=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if isMetaModel:
            self.isMetaModel.extend(isMetaModel)

        if isWritable:
            self.isWritable.extend(isWritable)

        if contents:
            self.contents.extend(contents)

        if domain is not None:
            self.domain = domain


class TextResource(FileResourceA):

    type = EAttribute(eType=FileResourceTypesE, derived=False,
                      changeable=True, default_value=FileResourceTypesE.TEXT)
    lines = EAttribute(eType=EString, derived=False, changeable=True, upper=-1)

    def __init__(self, *, type=None, lines=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if lines:
            self.lines.extend(lines)


class RawResource(FileResourceA):

    type = EAttribute(eType=FileResourceTypesE, derived=False,
                      changeable=True, default_value=FileResourceTypesE.RAW)
    data = EAttribute(eType=EByte, derived=False, changeable=True, upper=-1)

    def __init__(self, *, type=None, data=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if data:
            self.data.extend(data)


class ResourceSegment(RecursiveSegmetA):

    type = EAttribute(eType=SegmentTypesE, derived=False, changeable=True,
                      default_value=SegmentTypesE.RESOURCE)
    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value=':')

    def __init__(self, *, type=None, startCharacter=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if startCharacter is not None:
            self.startCharacter = startCharacter


class PathSegment(RecursiveSegmetA):

    type = EAttribute(eType=SegmentTypesE, derived=False,
                      changeable=True, default_value=SegmentTypesE.PATH)
    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value='/')
    feature = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, type=None, startCharacter=None, feature=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if startCharacter is not None:
            self.startCharacter = startCharacter

        if feature is not None:
            self.feature = feature


class ClassSegment(RecursiveSegmetA):

    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value='$')
    type = EAttribute(eType=SegmentTypesE, derived=False,
                      changeable=True, default_value=SegmentTypesE.CLAZZ)
    clazz = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, startCharacter=None, type=None, clazz=None, **kwargs):

        super().__init__(**kwargs)

        if startCharacter is not None:
            self.startCharacter = startCharacter

        if type is not None:
            self.type = type

        if clazz is not None:
            self.clazz = clazz


class InstanceOfSegment(RecursiveSegmetA):

    type = EAttribute(eType=SegmentTypesE, derived=False, changeable=True,
                      default_value=SegmentTypesE.INSTANCE)
    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value='?')
    clazz = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, type=None, startCharacter=None, clazz=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if startCharacter is not None:
            self.startCharacter = startCharacter

        if clazz is not None:
            self.clazz = clazz


class IdSegment(RecursiveSegmetA):

    type = EAttribute(eType=SegmentTypesE, derived=False,
                      changeable=True, default_value=SegmentTypesE.ID)
    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value='#')
    element = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, type=None, startCharacter=None, element=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if startCharacter is not None:
            self.startCharacter = startCharacter

        if element is not None:
            self.element = element


class MetaSegment(RecursiveSegmetA):

    type = EAttribute(eType=SegmentTypesE, derived=False,
                      changeable=True, default_value=SegmentTypesE.META)
    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value='@')
    identifierType = EAttribute(eType=MetaSegmentIdentifiersE, derived=False, changeable=True)

    def __init__(self, *, type=None, startCharacter=None, identifierType=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if startCharacter is not None:
            self.startCharacter = startCharacter

        if identifierType is not None:
            self.identifierType = identifierType


class HistorySegment(RecursiveSegmetA):

    type = EAttribute(eType=SegmentTypesE, derived=False, changeable=True,
                      default_value=SegmentTypesE.HISTORY)
    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value='!')
    element = EReference(ordered=True, unique=True, containment=False)

    def __init__(self, *, type=None, startCharacter=None, element=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if startCharacter is not None:
            self.startCharacter = startCharacter

        if element is not None:
            self.element = element


class ListOpSegment(LinearSegmetA):

    type = EAttribute(eType=SegmentTypesE, derived=False,
                      changeable=True, default_value=SegmentTypesE.LISTOP)
    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value='.')
    identifierType = EAttribute(eType=ListOpSegmentTypesE, derived=False, changeable=True)

    def __init__(self, *, type=None, startCharacter=None, identifierType=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if startCharacter is not None:
            self.startCharacter = startCharacter

        if identifierType is not None:
            self.identifierType = identifierType


class SelectorSegment(RecursiveSegmetA):

    type = EAttribute(eType=SegmentTypesE, derived=False, changeable=True,
                      default_value=SegmentTypesE.SELECTOR)
    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value='{')

    def __init__(self, *, type=None, startCharacter=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if startCharacter is not None:
            self.startCharacter = startCharacter


class IndexSegment(LinearSegmetA):

    type = EAttribute(eType=SegmentTypesE, derived=False,
                      changeable=True, default_value=SegmentTypesE.INDEX)
    startCharacter = EAttribute(eType=EChar, derived=False, changeable=True, default_value='[')

    def __init__(self, *, type=None, startCharacter=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if startCharacter is not None:
            self.startCharacter = startCharacter


@abstract
class LocalDomainA(DomainA, Directory, ActionHandlerA):

    baseDir = EAttribute(eType=EString, derived=False, changeable=True)
    baseDirAbs = EAttribute(eType=EString, derived=False, changeable=True)
    baseDirCwd = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, baseDir=None, baseDirAbs=None, baseDirCwd=None, **kwargs):

        super().__init__(**kwargs)

        if baseDir is not None:
            self.baseDir = baseDir

        if baseDirAbs is not None:
            self.baseDirAbs = baseDirAbs

        if baseDirCwd is not None:
            self.baseDirCwd = baseDirCwd



'''
Created on 24.06.2019

@author: Annighoefer
'''

from ..model import DomainA
from .. import CommandParser
from .. import ResultParser

import threading
from multiprocessing import Queue
import queue # imported for using queue.Empty exception

class MultiprocessingQueueDomainHost(DomainA):
    def __init__(self,domain,timeout=0.1):
        self.domain = domain
        self.cmdQueue = Queue()
        self.resQueue = Queue()
        self.timeout = timeout
        self.shallRun = True
        
    def Start(self):
        self.listenerThread = threading.Thread(target=self.__RemoteCommandHandler, args=())
        self.listenerThread.start()
        
    def Join(self):
        self.shallRun = False
        self.listenerThread.join()
        
    def __RemoteCommandHandler(self):
        while self.shallRun:
            try:
                cmdStr = self.cmdQueue.get(timeout=self.timeout)
                cmd = CommandParser.StringToCommand(cmdStr)
                res = self.domain.Do(cmd)
                resStr = ResultParser.ResultToString(res)
                self.resQueue.put(resStr) 
            except queue.Empty:
                continue
            except:
                break


'''
 Single Model MDB Provider
 Bjoern Annighoefer 2019
'''


from ...util import NoLogging
from ...event import MsgEvt,ChgTypes,EvtTypes
from .pyecoremdb import PyEcoreMdb

from pyecore.resources import ResourceSet, URI
from pyecore.ecore import EObject,EPackage

from timeit import default_timer
from threading import Timer
import traceback

class PyEcoreSingleFileMdbProvider():
    def __init__(self,modelfile,metamodelfile=None,saveTimeout=1.0,logger=NoLogging(),autoload=True):
        self.modelfile = modelfile
        self.metamodelfile = metamodelfile
        self.isLoaded = False
        self.rset = None
        self.modelroot = None
        
        self.logger = logger
        
        self.metamodelresource = None
        self.modelresource = None
        self.isDirty = False #indicates if the file needs to be saved
        self.saveTimeout = saveTimeout #This time is waited for any model modifications and than autosafe is called 
        self.saveTimer = None
        
        self.valueCodec = None #is set during coupling
        self.domain = None #is set during coupling
        self.mdb = None #is created later
        
        if(autoload):
            self.Load()
            
    def Close(self):
        #make sure any changes are saved
        #self.logger.Info("Closing single %s ..."%(self.modelfile)) #makes problems when main thread has ended already
        
        self.__StopDelayedSaving()
        self.__Save()
        
#     def Root(self):
#         return self.modelroot
    
#     def Class(self,packageName,className):
#         clazz = None
#         try:
#             pack = self.rset.metamodel_registry[packageName]
#             clazz = pack.getEClassifier(className)
#         except Exception as e:
#             raise EoqError(0,"Could not create class %s:%s: %s"%(packageName,className,str(e)))
#         return clazz
    
    def GetMdb(self):
        return self.mdb
        
    def Load(self,):
        self.logger.Info("Opening mdb for %s... "%(self.modelfile))
        start = default_timer()
        if(not self.isLoaded):
            self.rset = ResourceSet()
            if(self.metamodelfile):
                self.metamodelresource = self.rset.get_resource(URI(self.metamodelfile)) 
                metamodelroot = self.metamodelresource.contents[0]
                self.rset.metamodel_registry[metamodelroot.nsURI] = metamodelroot
                for child in metamodelroot.eAllContents():
                        if(isinstance(child,EPackage)):
                            self.rset.metamodel_registry[child.nsURI] = child
            self.modelresource = self.rset.get_resource(URI(self.modelfile)) 
            self.modelroot = self.modelresource.contents[0]  # We get the root (an EPackage here)
            self.mdb = PyEcoreMdb(self.modelroot,self.rset.metamodel_registry)
        end = default_timer()
        self.logger.Info("ok (%f s)"%(end-start))
        
    def CoupleWithDomain(self,domain,valueCodec):
        self.domain = domain
        self.valueCodec = valueCodec
        self.domain.Observe(self.OnChange,[EvtTypes.CHG])
        
    ''' sync model changes with workspace '''
        
    def OnChange(self,evts,src):
        for evt in evts:
            chg = evt
#             ctype = chg.a[1]
            target = self.valueCodec.Dec(chg.a[2]) #reverse what the domain did to the element
#             feature = chg.a[3]
#             oldOwner = self.valueCodec.Dec(chg.a[6])
            if(isinstance(target,EObject)):
                self.__SetDirty()
            
#     def Save(self):
#         #stop the current save action 
#         self.__StopDelayedSaving()
#         if(self.isDirty):
#             if(self.saveTimeout > 0):
#                 self.__StartDelayedSaving()
#             else:
#                 self.__Save() #skip threading for a 0 timeout
    
    ''' Saving functions '''
    def __SetDirty(self):
        self.isDirty = True
        self.__InitSave()
                      
    def __InitSave(self):
        #stop the current save action 
        self.__StopDelayedSaving()
        if(self.isDirty):
            if(self.saveTimeout > 0):
                self.__StartDelayedSaving()
            else:
                self.__Save() #skip threading for a 0 timeout
        
    def __StartDelayedSaving(self):
        self.saveTimer = Timer(self.saveTimeout,self.__DelayedSaving)
        self.saveTimer.start()
        
    def __StopDelayedSaving(self):
        if(self.saveTimer):
            self.saveTimer.cancel()
            self.saveTimer = None
            
    def __DelayedSaving(self):
        self.__Save()
        self.saveTimer = None
        
    def __Save(self):
        if(self.isDirty):
            evts = []
            try:
                self.__SaveResource()
                evts.append(MsgEvt("SingleFileMdb","Saved %s."%(self.modelfile)))
            except Exception as e:
                evts.append(MsgEvt("SingleFileMdb","Failed to save %s: %s"%(self.modelfile,str(e))))
                traceback.print_exc()
            #notify observers
            self.domain.NotifyObservers(evts, self)
        
    def __SaveResource(self):
        self.logger.Info("Saving %s ..."%(self.modelfile))
        start = default_timer()
        self.modelresource.save()
        self.isDirty = False
        end = default_timer()
        self.logger.Info("ok (%f s)"%(end-start))
    
#     def SetObjectDirty(self,obj):
#         #quit any previous saving timeout, this prevents doing the save during heavy DB interactions
#         self.__StopDelayedSaving()
#         #mark it dirty
#         self.isDirty = True
    


"""Definition of meta model 'xmlresourcemodel'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *


name = 'xmlresourcemodel'
nsURI = 'http://www.eoq.de/xmlresourcemodel/v1.0'
nsPrefix = 'de.eoq'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)


class Document(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, derived=False, changeable=True)
    version = EAttribute(eType=EString, derived=False, changeable=True)
    standalone = EAttribute(eType=EString, derived=False, changeable=True)
    encoding = EAttribute(eType=EString, derived=False, changeable=True)
    rootelement = EReference(ordered=True, unique=True, containment=True)
    comments = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, name=None, version=None, standalone=None, encoding=None, rootelement=None, comments=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if version is not None:
            self.version = version

        if standalone is not None:
            self.standalone = standalone

        if encoding is not None:
            self.encoding = encoding

        if rootelement is not None:
            self.rootelement = rootelement

        if comments:
            self.comments.extend(comments)


class Element(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, derived=False, changeable=True)
    content = EAttribute(eType=EString, derived=False, changeable=True)
    attributes = EReference(ordered=True, unique=True, containment=True, upper=-1)
    subelements = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, name=None, content=None, attributes=None, subelements=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if content is not None:
            self.content = content

        if attributes:
            self.attributes.extend(attributes)

        if subelements:
            self.subelements.extend(subelements)


class Attribute(EObject, metaclass=MetaEClass):

    name = EAttribute(eType=EString, derived=False, changeable=True)
    value = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, name=None, value=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if name is not None:
            self.name = name

        if value is not None:
            self.value = value


@abstract
class CommentA(EObject, metaclass=MetaEClass):

    content = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, content=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if content is not None:
            self.content = content


class Comment(CommentA):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)


class CData(CommentA):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)



'''
 Bjoern Annighoefer 2019
'''

from ..mdbaccessor import MdbAccessor
from ...command.command import CloModes
from ...util.error import EoqError

#Definition of pyecore based accessors
from pyecore.ecore import EPackage,EObject,EClass,EProxy,MetaEClass,\
    EClassifier
from pyecore.valuecontainer import EOrderedSet

import types
from itertools import chain


class PyEcoreMdbAccessor(MdbAccessor):
    def __init__(self,mdb=None,valueCodec=None):
        super().__init__()
        self.mdb = mdb #is necessary for creating classes by name
        self.vc = valueCodec
        ##make sure the root gets the primary encoding
        self.root = self.__ResolveAndEnc(self.mdb.Root())
        self.root_obj = self.mdb.Root()
        
        #tid
        self.latestTransactionId = 0
        
        #internal caching
        #self.allReferencesLut = {}
        
    def Lock(self):
        self.mdb.Lock()
    
    def Release(self):
        self.mdb.Release()
        
    ''' METAMODEL ACESSORS '''
   
    def GetAllMetamodels(self):
        ePackages = self.mdb.Metamodels()
        metamodels = [self.__ResolveAndEnc(m) for m in ePackages]
        return metamodels
     
    def RegisterMetamodel(self,package):
        ePackage = self.vc.Dec(package)
        self.mdb.AddMetamodel(None,ePackage) #name is retrieved internally from URI
        return True
    
    def UnregisterMetamodel(self,package):
        ePackage = self.vc.Dec(package)
        self.mdb.RemoveMetamodel(None,ePackage) #name is retrieved internally from URI
        return True

    ''' MODEL ACESSORS '''

    def GetRoot(self):
        root = self.mdb.Root()
        return self.__ResolveAndEnc(root)
        
    def Get(self,obj,feature):
        value = None
        eObj = self.vc.Dec(obj)
        eFeature = self.vc.Dec(feature)
        eValue = eObj.eGet(eFeature)
        #check if this was not a value but a method.
        if(isinstance(eValue,types.MethodType)): 
            eValue =  eValue()#call the method to get the values
        #convert any sets to lists
        if(isinstance(eValue, EOrderedSet) or 
           isinstance(eValue, set) or 
           isinstance(eValue, list)):
            value = self.__RemoveNullRefs([ self.__ResolveAndEnc(v) for v in eValue])
        else:
            value = self.__ResolveAndEnc(eValue)
        return value
    
    def Set(self,obj,feature,value):
        eObj = self.vc.Dec(obj)
        eFeature = self.vc.Dec(feature)
        oldEValue = eObj.eGet(eFeature)
        eValue = self.vc.Dec(value)
        oldValue = self.__ResolveAndEnc(oldEValue)
        #preserve the old values
        oldOwner = None
        oldFeature = None
        oldIndex = None
        eOldOwner = None
        if(isinstance(eValue, EObject)):
            eOldOwner = eValue.eContainer()
            if(eOldOwner):
                oldOwner = self.__ResolveAndEnc(eOldOwner)
                eOldFeature = eValue.eContainmentFeature()
                oldFeature = self.__ResolveAndEnc(eOldFeature)
                if(eOldFeature.many):
                    oldIndex = eOldOwner.eGet(eOldFeature).index(eValue)
        eObj.eSet(eFeature,eValue)
        return (oldValue,oldOwner,oldFeature,oldIndex)
    
    def Add(self,obj,feature,child):
        eObj = self.vc.Dec(obj)
        eFeature = self.vc.Dec(feature)
        eChild = self.vc.Dec(child)
        oldEValue = eObj.eGet(eFeature)
        #preserve the old values
        oldOwner = None
        oldFeature = None
        oldIndex = None
        eOldOwner = None
        if(isinstance(eChild, EObject)):
            eOldOwner = eChild.eContainer()
            if(eOldOwner):
                oldOwner = self.__ResolveAndEnc(eOldOwner)
                eOldFeature = eChild.eContainmentFeature()
                oldFeature = self.__ResolveAndEnc(eOldFeature)
                if(eOldFeature.many):
                    oldIndex = eOldOwner.eGet(eOldFeature).index(eChild)
        #this must happen before in order to copy the old value
        oldValue = [ self.__ResolveAndEnc(v) for v in oldEValue]
        #add the new child
        oldEValue.add(eChild)
        return (oldValue,oldOwner,oldFeature,oldIndex)
    
    def Remove(self,obj,feature,child):
        eObj = self.vc.Dec(obj)
        eFeature = self.vc.Dec(feature)
        eChild = self.vc.Dec(child)
        oldEValue = eObj.eGet(eFeature)
        #preserve the old values
        oldOwner = None
        oldFeature = None
        oldIndex = None
        eOldOwner = None
        if(isinstance(eChild, EObject)):
            eOldOwner = eChild.eContainer()
            if(eOldOwner):
                oldOwner = self.__ResolveAndEnc(eOldOwner)
                eOldFeature = eChild.eContainmentFeature()
                oldFeature = self.__ResolveAndEnc(eOldFeature)
                if(eOldFeature.many):
                    oldIndex = eOldOwner.eGet(eOldFeature).index(eChild)
        #this must happen before in order to copy the old value
        oldValue = [ self.__ResolveAndEnc(v) for v in oldEValue]
        #remove the child
        eChild = self.vc.Dec(child)
        oldEValue.remove(eChild)
        return (oldValue,oldOwner,oldFeature,oldIndex)
    
    def Move(self,obj,newIndex):
        eObj = self.vc.Dec(obj)
        eContainer = eObj.eContainer()
        eFeature = eObj.eContainmentFeature()
        
        oldEValue = eContainer.eGet(eFeature)
        #preserve the old values
        oldOwner = self.__ResolveAndEnc(eContainer)
        oldFeature = self.__ResolveAndEnc(eFeature)
        oldIndex = eContainer.eGet(eFeature).index(eObj)
        #this must happen before in order to copy the old value
        oldValue = [ self.__ResolveAndEnc(v) for v in oldEValue]
        
        #do the move if it makes sense
        if(newIndex!=oldIndex):
            oldEValue.pop(oldIndex)
            oldEValue.insert(newIndex,eObj)
        return (oldValue,oldOwner,oldFeature,oldIndex)
        
    
    def Clone(self, obj, mode):
        clone = None
        eObj = self.vc.Dec(obj)
        if(CloModes.CLS==mode):
            clone = self.__ECloneClass(eObj)
        elif(CloModes.ATT==mode):
            clone = self.__ECloneAttributes(eObj)
        elif(CloModes.FLT==mode):
            clone = self.__ECloneFlat(eObj)
        elif(CloModes.FUL==mode):
            clone = self.__ECloneFull(eObj)
        elif(CloModes.DEP==mode):
            clone = self.__ECloneDeep(eObj)
        return self.__ResolveAndEnc(clone)
    
    def Create(self, clazz,n,constructorArgs=[]):
        eClass = self.vc.Dec(clazz)
        eConstructorArgs = [self.vc.Dec(a) for a in constructorArgs] # necessary if this contains an eObject
        if(n==1):
            newObjects = self.__ResolveAndEnc(self.__CreateEClassInstance(eClass,eConstructorArgs))
        else:
            newObjects = [self.__ResolveAndEnc(self.__CreateEClassInstance(eClass,eConstructorArgs)) for i in range(n)]
        return newObjects
        
    def CreateByName(self,packageName,className,n,constructorArgs=[]):
        newObjects = None
        eConstructorArgs = [self.vc.Dec(a) for a in constructorArgs] # necessary if this contains an eObject
        try:
            eClass = self.__GetClassByName(packageName, className)
            if(n==1):
                newObjects = self.__ResolveAndEnc(self.__CreateEClassInstance(eClass,eConstructorArgs))
            else:
                newObjects = [self.__ResolveAndEnc(self.__CreateEClassInstance(eClass,eConstructorArgs)) for i in range(n)]
        except Exception as e:
            raise EoqError(0,"Could not create class %s:%s: %s"%(packageName,className,str(e)))
        return newObjects
    
    def GetClassByName(self,packageName,className):
        eClass = self.__GetClassByName(packageName, className)
        return self.__ResolveAndEnc(eClass)
    
    def AddConstraint(self,constraint):
        index = None
        eClass = self.__GetClassByName('http://www.example.org/constraintmodel', 'Constraint')
        eClass_instance = self.__CreateEClassInstance(eClass)
        eClass_instance.eSet('constraintType',constraint.constraintType)
        eClass_instance.eSet('name',constraint.name)
        eClass_instance.eSet('target',constraint.target)
        eClass_instance.eSet('feature',constraint.feature)
        eClass_instance.eSet('law',constraint.law)
        eClass_instance.eSet('annotation',constraint.annotation)
        for i in range(len(self.root_obj.eContents)):
            if self.root_obj.eContents[i].name == 'cdb.constraintmodel':
                index = i
        if not(index == None):
            self.root_obj.eContents[index].eContents[0].constraint.append(eClass_instance)
            print("Constraint is added")
        
    def RemoveConstraint(self,constraint):
        index = None
        eClass = self.__GetClassByName('http://www.example.org/constraintmodel', 'Constraint')
        eClass_instance = self.__CreateEClassInstance(eClass)
        eClass_instance.eSet('constraintType',constraint.constraintType)
        eClass_instance.eSet('name',constraint.name)
        eClass_instance.eSet('target',constraint.target)
        eClass_instance.eSet('feature',constraint.feature)
        eClass_instance.eSet('law',constraint.law)
        eClass_instance.eSet('annotation',constraint.annotation)
        for i in range(len(self.root_obj.eContents)):
            if self.root_obj.eContents[i].name == 'cdb.constraintmodel':
                index = i
        if not(index == None):
            for eobj in self.root_obj.eContents[index].eContents[0].eContents:
                if str(eobj.constraintType) == str(eClass_instance.constraintType) and eobj.name == eClass_instance.name and eobj.target == eClass_instance.target and eobj.feature == eClass_instance.feature and eobj.law == eClass_instance.law and eobj.annotation == eClass_instance.annotation:
                    eobj.delete()
            
    def CalculateSimilarityIndex(self, obj, old):
        '''
        Calculates a Similarity Index of all children of "old" with respect to "obj".
        "Obj" is from "new" with level n ...
        ... and "old" the corresponding object from "old" with level n+1
        Calls 4 Subfunctions: Name, Attributes, References, Index/Position
        which can award up to 0.25 points each -> Maximum: 1, Minimum 0
        '''
        allSI      = [] # stores the SI for each child of old
        allChanges = [] # stores changes for MergeRoutine to handle
        eObj    = self.vc.Dec(obj)
        typeObj = eObj.eClass.name
        childrenOld = self.GetAllChildrenOfType(old, typeObj)
        # print('# STARTING CSI FOR', obj, 'and', childrenOld)
        for childOld in childrenOld:
            # print('### CSI LOOP FOR', obj, 'and', childOld)
            childObj = self.vc.Dec(childOld)
            childType = childObj.eClass.name
            SI = 0.0
            changes = []
            ### 1 NAME
            dSIName = self.__SimilarityIndexName(obj, childOld)
            # print('### 1 NAME:', dSI)
            if not dSIName == 0.25:
                changes.append('N')
            ### 2 REFERENCES
            dSIRef = self.__SimilarityIndexReferences(obj, childOld)
            # print('### 2 REFERENCES:', dSI)
            if not dSIRef == 0.25:
                changes.append('R')
            # 3 ATTRIBUTES
            dSIAttr = self.__SimilarityIndexAttributes(obj, childOld)
            # print('### 3 ATTRIBUTES:', dSI)
            if not dSIAttr == 0.25:
                changes.append('A')
            # 4 INDEX
            dSIIndex = self.__SimilarityIndexPosition(obj, childOld)
            # print('### 4 INDEX:', dSI)
            if not dSIIndex == 0.25:
                changes.append('I')
            # append to SI array rounded
            SI = dSIName + dSIRef + dSIAttr + dSIIndex
            # print('### End of CSI Loop:', dSIName, dSIRef, dSIAttr, dSIIndex)
            allSI.append(round(SI, 2))
            allChanges.append(changes)
            # if SI >= 0.99: # we found an identical object, so no need to continue
            #     break
        # return all SI and the corresponding objects
        return [allSI, childrenOld, allChanges]

    ''' CLASS ACESSORS '''
    
    def Class(self,obj):
        eObj = self.vc.Dec(obj)
        res = self.__ResolveAndEnc(eObj.eClass)
        return res
    
    def ClassName(self,obj):
        eObj = self.vc.Dec(obj)
        res = eObj.eClass.name
        return res
    
    def Package(self,clazz):
        res = None
        eClass = self.vc.Dec(clazz)
        if(isinstance(eClass,EClassifier)):
            res = self.__ResolveAndEnc(eClass.eContainer())
        else:
            raise EoqError(0,"Package works only for class objects")
        return res
    
    def Supertypes(self,clazz):
        res = []
        eClass = self.vc.Dec(clazz)
        #find the root package
        if(isinstance(eClass,EClass)):
            res = [self.__ResolveAndEnc(s) for s in eClass.eSuperTypes]
        else:
            raise EoqError(0,"Supertypes works only for class objects")
        return res
    
    def AllSupertypes(self,clazz):
        res = []
        eClass = self.vc.Dec(clazz)
        #find the root package
        if(isinstance(eClass,EClass)):
            res = [self.__ResolveAndEnc(s) for s in eClass.eAllSuperTypes()]
        else:
            raise EoqError(0,"All Supertypes works only for class objects")
        return res
    
    def Implementers(self,clazz):
        res = []
        eClass = self.vc.Dec(clazz)
        if(isinstance(eClass,EClass)):
            #find the root package
            rootPackage = eClass.eContainer()
            container = rootPackage.eContainer()
            while(isinstance(container,EPackage)):
                rootPackage = container
                container = rootPackage.eContainer()
            #find the implementers 
            res = [self.__ResolveAndEnc(c) for c in rootPackage.eAllContents() if ((type(c) == EClass) and (eClass in c.eSuperTypes))]
        else:
            raise EoqError(0,"Implementers works only for class objects")
        return res
    
    def AllImplementers(self,clazz):
        res = []
        eClass = self.vc.Dec(clazz)
        if(isinstance(eClass,EClass)):
            #find the root package
            rootPackage = eClass.eContainer()
            container = rootPackage.eContainer()
            while(isinstance(container,EPackage)):
                rootPackage = container
                container = rootPackage.eContainer()
            #find the implementers 
            res = [self.__ResolveAndEnc(c) for c in rootPackage.eAllContents() if ((type(c) == EClass) and (eClass in c.eAllSuperTypes()))]
        else:
            raise EoqError(0,"All implementers works only for class objects")
        return res
    
    def GetParent(self,obj):
        res = self.__ResolveAndEnc(self.vc.Dec(obj).eContainer())
        return res
    
    def GetAllParents(self,obj):
        res = []
        eObj = self.vc.Dec(obj)
        parent = eObj.eContainer()
        while(parent):
            res.append(self.__ResolveAndEnc(parent))
            parent = parent.eContainer()
        res.reverse() #reverse from earliest anchestor to the child
        return res
    
    def GetAssociates(self,obj,root):
        eRes =  []
        eObj = self.vc.Dec(obj)
        eRoot = self.vc.Dec(root)
        
        allReferencesLut = {}
        
        eChildsOfRoot = eRoot.eAllContents()
        for eChild in eChildsOfRoot:
            eClass = eChild.eClass
            try: 
                #eReferences = self.allReferencesLut[eClass]
                eReferences = allReferencesLut[eClass]
            except KeyError:
                eReferences = [r for r in eChild.eClass.eAllReferences() if not r.containment]
                #self.allReferencesLut[eClass] = eReferences
                allReferencesLut[eClass] = eReferences
            for r in eReferences:
                if(isinstance(eObj,r.eType)):
                    rVal = eChild.eGet(r.name)
                    if r.upperBound==1: #single ref
                        #rObj = self.__ResolveProxyAndBuildInClasses(rVal)
                        rObj = self.mdb.ResolveProxy(rVal)
                        if(eObj == rObj):
                            eRes.append(eChild)
                    else:
                        rObjs = rVal
                        for rObj in rObjs:
                            #rObj = self.__ResolveProxyAndBuildInClasses(rObj)
                            rObj = self.mdb.ResolveProxy(rObj)
                            if(eObj == rObj):
                                eRes.append(eChild)                           
        return self.__RemoveNullRefs([self.__ResolveAndEnc(o) for o in eRes])
    
    def GetIndex(self,obj):
        eObj = self.vc.Dec(obj)
        res = 0
        f = eObj.eContainmentFeature()
        if(None==f or f.upperBound == 1):
            res = None ## denotes the index makes no sense for single object references
        else: #index must be determined by backwards search
            #this backward search seems not optimal
            containingIndex = 0 #Eoq indicies start at 0
            container = eObj.eContainer()
            for sibling in container.eGet(f):
                if(sibling==eObj):
                    res = containingIndex
                    break;
                containingIndex = containingIndex+1
        return res
    
    def GetContainingFeature(self,obj):
        eObj = self.vc.Dec(obj)
        res = None
        f = eObj.eContainmentFeature()
        if(f):
            res = self.__ResolveAndEnc(f)
        else:
            res = None
        return res
    
    #def IsObject(self,value,encoder):
    #    return isinstance(encoder.decode(value),EObject)
        
    
    def GetAllChildrenOfType(self,obj,clazz):
        res = []
        eObj = self.vc.Dec(obj)
        childs = eObj.eAllContents()
        searchlist = chain(iter([eObj]),childs)
        if(isinstance(clazz,str)):
            className = clazz
            res = [self.__ResolveAndEnc(e) for e in searchlist if self.__IsType(e,className)]
        else:
            eClass = self.vc.Dec(clazz)
            res = [self.__ResolveAndEnc(e) for e in searchlist if e.eClass == eClass]
        return res
    
    def GetAllChildrenInstanceOfClass(self,obj,clazz):
        res = []
        eObj = self.vc.Dec(obj)
        childs = eObj.eAllContents()
        searchlist = chain(iter([eObj]),childs)
        if(isinstance(clazz,str)):
            className = clazz
            res = [self.__ResolveAndEnc(e) for e in searchlist if self.__IsInstanceOf(e,className)]
        else:
            eClass = self.vc.Dec(clazz)
            res = [self.__ResolveAndEnc(e) for e in searchlist if isinstance(e, eClass)]
        return res
    
    def GetAllFeatures(self,obj):
        res = [self.__ResolveAndEnc(x) for x in self.vc.Dec(obj).eClass.eAllStructuralFeatures()]
        return res
    
    def GetAllFeatureNames(self,obj):
        res = [x.name for x in self.vc.Dec(obj).eClass.eAllStructuralFeatures()]
        return res
    
    def GetAllFeatureValues(self,obj):
        eObj = self.vc.Dec(obj)
        res = [((self.__ResolveAndEnc(eObj.eGet(x)) if (x.upperBound==1) else self.__RemoveNullRefs([self.__ResolveAndEnc(c) for c in eObj.eGet(x)])))  for x in eObj.eClass.eAllStructuralFeatures()]
        return res
    
    def GetAllAttributes(self,obj):
        eObj = self.vc.Dec(obj)
        res = [self.__ResolveAndEnc(x) for x in eObj.eClass.eAllAttributes()]
        return res
    
    def GetAllAttributeNames(self,obj):
        eObj = self.vc.Dec(obj)
        res = [x.name for x in eObj.eClass.eAllAttributes()]
        return res
    
    def GetAllAttributeValues(self,obj):
        eObj = self.vc.Dec(obj)
        attributes = [x for x in eObj.eClass.eAllAttributes()]
        res = [ self.__ResolveAndEnc(eObj.eGet(a)) if (a.upperBound==1) else [self.__ResolveAndEnc(c) for c in eObj.eGet(a)] for a in attributes]
        return res
    
    def GetAllReferences(self,obj):
        eObj = self.vc.Dec(obj)
        res = [self.__ResolveAndEnc(x) for x in eObj.eClass.eAllReferences()]
        return res
    
    def GetAllReferenceNames(self,obj):
        eObj = self.vc.Dec(obj)
        res = [x.name for x in eObj.eClass.eAllReferences()]
        return res
    
    def GetAllReferenceValues(self,obj):
        eObj = self.vc.Dec(obj)
        references = [x for x in eObj.eClass.eAllReferences()]
        res = [ self.__ResolveAndEnc(eObj.eGet(a)) if (a.upperBound==1) else self.__RemoveNullRefs([self.__ResolveAndEnc(c) for c in eObj.eGet(a)]) for a in references]
        return res
    
    def GetAllContainments(self,obj):
        eObj = self.vc.Dec(obj)
        res = [self.__ResolveAndEnc(x) for x in eObj.eClass.eAllReferences() if x.containment]
        return res
    
    def GetAllContainmentNames(self,obj):
        eObj = self.vc.Dec(obj)
        res = [x.name for x in eObj.eClass.eAllReferences() if x.containment]
        return res
    
    def GetAllContainmentValues(self,obj):
        eObj = self.vc.Dec(obj)
        containments = [x for x in eObj.eClass.eAllReferences() if x.containment]
        #res = [ [self.__ResolveAndEnc(c) for c in eObj.eGet(a)] for a in containments if (a.upperbound>1) else valueCodec(eObj.eGet(a))]
        res = [ (self.__ResolveAndEnc(eObj.eGet(a)) if (a.upperBound==1) else [self.__ResolveAndEnc(c) for c in eObj.eGet(a)]) for a in containments]  
        return res
    
    '''
        PRIVATE METHODS
    '''
   
    def _OnMdbEvents(self,evts,src):
        self.NotifyObservers(evts, src)
        
    def __GetClassByName(self,packageName,className):
        try:
            pack = self.mdb.GetMetamodel(packageName)
            eClass = pack.getEClassifier(className)
        except:
            raise EoqError(0,"Unknown package %s or class %s."%(packageName,className))
        return eClass
        
    def __CreateEClassInstance(self,eClass,constructorArgs=[]):
        eObj = None
        if(0<len(constructorArgs)):
            eObj = eClass(*constructorArgs)
        else:
            eObj = eClass()
        return eObj
   
    def __IsType(self,obj,className):
        return (obj.eClass.name==className)
   
    def __IsInstanceOf(self,obj,className):
        result = self.__IsType(obj,className)
        if(not result):
            for superType in obj.eClass.eAllSuperTypes():
                if(superType.name==className):
                    result = True
                    break
            if('EObject'==className):
                result = isinstance(obj,EObject)
        return result
    
    def __ECloneClass(self,eObj):
        clazz = eObj.eClass
        clone = clazz()
            
        return clone
    
    def __ECloneAttributes(self,eObj):
        clone = self.__ECloneClass(eObj)
        clazz = eObj.eClass
        attributes = clazz.eAllAttributes()
        for attribute in attributes:
            if(attribute.many):
                for attributeValue in eObj.eGet(attribute):
                    clone.eGet(attribute).add(attributeValue)
            else:
                clone.eSet(attribute,eObj.eGet(attribute))    
        return clone
    
    def __ECloneFlat(self,eObj):
        clone = self.__ECloneAttributes(eObj)
        clazz = eObj.eClass
        references = clazz.eAllReferences()
        for reference in references:
            if(not reference.containment):
                if(reference.many):
                    for refObj in eObj.eGet(reference):
                        clone.eGet(reference).add(refObj)
                else:
                    clone.eSet(reference,eObj.eGet(reference))
        return clone
    
    def __ECloneDeep(self,eObj,copyReferences=True):
        clone = self.__ECloneAttributes(eObj)
        clazz = eObj.eClass
        references = clazz.eAllReferences()
        for reference in references:
            if(reference.containment):
                if(reference.many):
                    for child in eObj.eGet(reference):
                        clonedChild = self.__ECloneDeep(child,copyReferences)
                        clone.eGet(reference).add(clonedChild)
                else:
                    child = eObj.eGet(reference)
                    if(child): #only clone non empty single containments
                        clonedChild = self.__ECloneDeep(child,copyReferences)
                        clone.eSet(reference,clonedChild)
            elif(copyReferences): #non containments
                if(reference.many):
                    for refObj in eObj.eGet(reference):
                        clone.eGet(reference).add(refObj)
                else:
                    clone.eSet(reference,eObj.eGet(reference))
            
        return clone
    
    def __ECloneFull(self,eObj):
        clone = self.__ECloneDeep(eObj,False)
        #clazz = eObj.eClass
        self.__ECloneFullReferenceUpdater(eObj,clone,eObj,clone)            
        return clone
    
    def __ECloneFullReferenceUpdater(self,eObj,clonedEObject,root,clone):
        clazz = eObj.eClass
        references = clazz.eAllReferences()
        for reference in references:
            if(reference.containment):
                if(reference.many):
                    nChilds = len(eObj.eGet(reference))
                    for i in range(nChilds):
                        child = eObj.eGet(reference)[i]
                        clonedChild = clonedEObject.eGet(reference)[i]
                        self.__ECloneFullReferenceUpdater(child, clonedChild, root, clone)
                else:
                    child = eObj.eGet(reference)
                    if(child): #only clone non empty single containments
                        clonedChild = clonedEObject.eGet(reference)
                        self.__ECloneFullReferenceUpdater(child, clonedChild, root, clone)
            else: #non containments
                if(reference.many):
                    for oldRef in eObj.eGet(reference):
                        newRef = self.__ECloneFullFindCorrespondingReference(oldRef,root,clone)
                        clonedEObject.eGet(reference).add(newRef)
                else:
                    oldRef = eObj.eGet(reference)
                    if(oldRef): #null references need no cloning
                        newRef = self.__ECloneFullFindCorrespondingReference(oldRef,root,clone)
                        clonedEObject.eSet(reference,newRef)
        return clone
    
    def __ECloneFullFindCorrespondingReference(self,oldRef,root,clone):
        newRef = oldRef
        if(self.__ECloneFullIsParentOf(oldRef,root)): #we only need an new reference if the root contains the referred object
            featurePath = self.__ECloneFullGetPathFromTo(root,oldRef)
            newRef = clone
            for p in featurePath: #p is a tuple (featureName, many, index)
                feature = p[0]
                if(p[1]): #many
                    index = p[2]
                    newRef = newRef.eGet(feature)[index]
                else: #single
                    newRef = newRef.eGet(feature)
        return newRef
    
    def __ECloneFullIsParentOf(self,child,possibleParent):
        if(child == possibleParent) :
            return True
        parent = child.eContainer()
        if(parent and isinstance(parent, EObject)):
            return self.__ECloneFullIsParentOf(parent, possibleParent)
        return False #has no parent any more, return false
    
    def __ECloneFullGetPathFromTo(self,anchestor,child):
        featurePath = []
        if(anchestor!=child):
            parent = child.eContainer()
            if(parent and isinstance(parent, EObject)):
                feature = child.eContainmentFeature()
                featureName = feature.name
                many = feature.many
                index = 0
                if(many):
                    index = parent.eGet(featureName).index(child)
                featurePath = [(featureName,many,index)]
                featurePath = self.__ECloneFullGetPathFromTo(anchestor,parent) + featurePath
        return featurePath #has no parent any more, return false
    
#     def __ResolveProxyAndBuildInClasses(self,obj): 
#         if(isinstance(obj, EProxy)):
#             try: #resolve can fail if files have been modified or objects been deleted before the proxy has been resolved
#                 obj.force_resolve()
#                 obj = obj._wrapped #remove the outer proxy
#             except Exception as e:
#                 traceback.print_exc()
#                 print("WARN: Unresolvable proxy found and removed")
#                 obj = None;     
#         if(isinstance(obj, (MetaEClass,type,types.ModuleType))): #this is necessary to mask compiled model instances
#             obj = obj.eClass
#         return obj
#     
    def __ResolveAndEnc(self,obj):
        return self.vc.Enc(self.mdb.ResolveProxy(obj))
    
    def __RemoveNullRefs(self,objs):
        return list(filter(None,objs))
    
    def __SimilarityIndexName(self, obj, childOld):
        dSI = 0.0
        eObj     = self.vc.Dec(obj)
        childObj = self.vc.Dec(childOld)
        # is there a name attribute and is it identical? SI += 0.25
        if hasattr(childObj,'name') and childObj.name:
            if childObj.name == eObj.name:
                dSI = 0.25
        # nameless objects need to get + 0.25 if eObj also has no name
        elif (not hasattr(childObj, 'name') or not childObj.name) and (not hasattr(eObj, 'name') or not eObj.name):
            dSI = 0.25
        return dSI


    def __SimilarityIndexReferences(self, obj, childOld):
        dSI = 0.0
        eObj     = self.vc.Dec(obj)
        childObj = self.vc.Dec(childOld)
        referencesObj          = self.GetAllReferences(obj)
        referenceNamesObj      = self.GetAllReferenceNames(obj)
        referenceValuesObj     = self.GetAllReferenceValues(obj)
        # collect the number of valid referenced objects (obj)
        nValidRefValues = 0
        for ref in referenceValuesObj:
            if isinstance(ref, list):
                if len(ref) > 0:
                    nValidRefValues += 1
            else:
                nValidRefValues += 1
        # collect the names of the referenced objects (childOld)
        referencesChildOld      = self.GetAllReferences(childOld)
        referenceNamesChildOld  = self.GetAllReferenceNames(childOld)
        referenceValuesChildOld = self.GetAllReferenceValues(childOld)
        # iterate all references of obj (from here on abbreviated by F)
        # print('-- SIRef: Old:', childOld, 'New:', obj)
        for refF, refNameF, refValueF in zip(referencesObj, referenceNamesObj, referenceValuesObj):
            # match the reference value
            # print('- Trying to match reference:', refNameF, 'with values:', refValueF)
            matchedReferenceName = False
            # iterate all references of childOld and find the corresponding reference
            for i, [refO, refNameO, refValueO] in enumerate(zip(referencesChildOld, referenceNamesChildOld, referenceValuesChildOld)):
                if refNameO == refNameF:
                    matchedReferenceName = True
                    # print('--- Matched:', refNameO, refValueO)
                    # now i is the index of the matched reference
                    break
            # everything needs to be in lists
            if not isinstance(referenceValuesObj[i], list):
                referenceValuesObj[i] = [referenceValuesObj[i]]
            if not isinstance(referenceValuesChildOld[i], list):
                referenceValuesChildOld[i] = [referenceValuesChildOld[i]]
            # if the reference is matched and both have referenced objects, ...
            # ... we can compare the referenced objects' names
            if (matchedReferenceName 
                and len(referenceValuesObj[i]) > 0 
                and len(referenceValuesChildOld[i]) > 0):
                matchcounter = 0
                for objF in referenceValuesObj[i]:
                    decObjF = self.vc.Dec(objF)
                    if hasattr(decObjF, 'name') and decObjF.name:
                        nameF = decObjF.name
                    else:
                        nameF = 'noname'
                    matchedReference = False
                    for j, objO in enumerate(referenceValuesChildOld[i]):
                        # decode and compare
                        decObjO = self.vc.Dec(objO)
                        if hasattr(decObjO, 'name') and decObjO.name:
                            nameO = decObjO.name
                        else:
                            nameO = 'noname'
                        # see if we matched the reference
                        if nameO == nameF:
                            matchedReference = True
                            break
                    if matchedReference:
                        # print('--- Matched objects.  ',
                        # 	'Reference:', refNameF, 'Old:', nameO, 'New:', nameF)
                        matchcounter += 1
                    else:
                        pass
                        # print('--- Unmatched objects.',
                        # 	'Reference:', refNameF, 'Old:', nameO, 'New:', nameF)
                # print('- End of object comparison. Matchcounter:', matchcounter,
                #     'number of objects from this reference:',
                #     len(referenceValuesObj[i]))
                dSI += (matchcounter/len(referenceValuesObj[i]))*0.25/nValidRefValues
        if not any(referenceValuesObj) and not any(referenceValuesChildOld):
            # if both do not have references -> they have identical references
            dSI = 0.25
        return round(dSI, 2)

    def __SimilarityIndexAttributes(self, obj, childOld):
        dSI = 0
        eObj     = self.vc.Dec(obj)
        childObj = self.vc.Dec(childOld)
        attributeNamesObj  = self.GetAllAttributeNames(obj)
        attributeValuesObj = self.GetAllAttributeValues(obj)
        if 'name' in attributeNamesObj:
            attributeValuesObj.pop(attributeNamesObj.index('name'))
            attributeNamesObj.remove('name')
        attributeNamesChildOld  = self.GetAllAttributeNames(childOld)
        attributeValuesChildOld = self.GetAllAttributeValues(childOld)
        if 'name' in attributeNamesChildOld:
            attributeValuesChildOld.pop(attributeNamesChildOld.index('name'))
            attributeNamesChildOld.remove('name')
        nAttrs = len(attributeNamesObj)
        for (nameF, valueF) in zip(attributeNamesObj, attributeValuesObj):
            for (nameO, valueO) in zip(attributeNamesChildOld, attributeValuesChildOld):
                found = False
                if nameF == nameO:
                    if valueF == valueO:
                        dSI += 0.25/nAttrs
                        found = True
                        break
        if not attributeNamesObj and not attributeNamesChildOld:
            # if both do not have attributes -> they have identical attributes
            dSI = 0.25
        return round(dSI, 2)

    def __SimilarityIndexPosition(self, obj, childOld):
        dSI = 0
        eObj     = self.vc.Dec(obj)
        childObj = self.vc.Dec(childOld)
        if self.GetIndex(childObj) == self.GetIndex(eObj):
            dSI += 0.25
        return dSI
    
